// -*- C++ -*-
/*
 * CL Open EVSE Firmware
 *
 * Copyright (c) 2013-2014 Sam C. Lin <lincomatic@gmail.com>
 *
 * This file is part of CL Open EVSE.

 * CL Open EVSE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.

 * CL Open EVSE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with CL Open EVSE; see the file COPYING.  If not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <avr/pgmspace.h>
#include "rapi_proc.h"


// convert 2-digit hex string to uint8_t
uint8_t htou8(const char *s){
  uint8_t u = 0;
  for (int i=0;i < 2;i++) {
    if (i == 1) u <<= 4;
    char c = s[i];
    if ((c >= '0') && (c <= '9')) {
      u += c - '0'; }
    else if ((c >= 'A') && (c <= 'F')) {
      u += c - 'A' + 10; }}
  return u; }

// convert decimal string to uint32_t
uint32_t dtou32(const char *s) {
  uint32_t u = 0;
  while (*s) {
    u *= 10;
    u += *(s++) - '0'; }
  return u; }

EvseRapiProcessor::EvseRapiProcessor() {}

void EvseRapiProcessor::init() {
  pinMode(PIN_DE_RE,OUTPUT);   // PA6 to DE and /RE (Max 485)
  digitalWrite(PIN_DE_RE,LOW); // Listen Receiver enabled
  delay(400);           // give I2C devices time to be ready before running
                        // code that wants to initialize I2C devices.
                        //  Otherwise a hang can occur upon powerup.  
  reset();
  Serial1.begin(250000);        // Hardware Serial Teensy2 on Pin PD2 (Rx) and PD3 (Tx)  
  Serial.begin(115200);         // Serial monitor
  Serial.println("CLEVER simulator"); // Hardware Serial Teensy2 on Pin PD2 (Rx) and PD3 (Tx)  
  write("CLEVER simulator\n");        // Hardware Serial Teensy2 on Pin PD2 (Rx) and PD3 (Tx)  
  setTime(12, 00, 00, 01, 06, 20);
  Serial.flush();
  Serial1.flush();
}

int EvseRapiProcessor::doCmd() {  
  int rc = 1;
  char idH = 'N';
  char idL = 'N';
  int bcnt = available(); // characters in serial buffer available  
  if (bcnt) {    
    for (int i=0; i < (bcnt); i++) {         // treat all characters
      char c = read();
      if (echo) write(c);
      if (c == ESRAPI_SOC) {                 // if start of command sign
        buffer[0] = ESRAPI_SOC;              // write in in the buffer
        bufCnt = 1;
        cmdIdCnt = 1;        
      }
      else if ((cmdIdCnt==1) && (c == EVCS_BUS)) { // check IDH (bus)
        idH = EVCS_BUS;
        cmdIdCnt=2;                    
      }
      else if ((cmdIdCnt==2) && (c == EVCS_ID)) {  // check IDL (number on bus)
        idL = EVCS_ID;
        cmdIdCnt=0;                    
      }
      else if ((buffer[0] == ESRAPI_SOC) && (idH == EVCS_BUS) && (idL == EVCS_ID)) {          
        if (bufCnt < ESRAPI_BUFLEN) {               // check if no buffer overflow
          if (c == ESRAPI_EOC) { 
            buffer[bufCnt++] = 0;
            if (!tokenize()) {              
              rc = processCmd();
            }
	          else {
	            reset();
	            response(0);
	          }
	        }
	        else { 
	          buffer[bufCnt++] = c;
	        }
	      } 
	      else { // buffer overflow
	        reset();
	      }
      }
    }
  } // nothing in buffer  
  return rc;
}

void EvseRapiProcessor::sendEvseState() {
  sprintf(g_sTmp,"%c%c%cST %02x%c",ESRAPI_SOC,EVCS_BUS,EVCS_ID,EVCS_STATE_A,ESRAPI_EOC);
  write(g_sTmp);
  Serial1.write(g_sTmp);
}

void EvseRapiProcessor::sendRFID_NUID() {
  /*char HEXH = (i >> 4) + 0x30;
  if (HEXH > 0x39) HEXH +=7;
  char HEXL = (i & 0x0F) + 0x30;
  if (HEXL > 0x39) HEXL +=7;
  m_Lcd.setCursor(x,y);
  m_Lcd.write(HEXH); // write for char!!
  m_Lcd.setCursor(x+1,y);
  m_Lcd.write(HEXL);*/

  sprintf(g_sTmp,"%c%c%cRF %02x%02x%02x%02x%c",ESRAPI_SOC,EVCS_BUS,EVCS_ID,1,2,5,7,ESRAPI_EOC);
  write(g_sTmp);
}


int EvseRapiProcessor::tokenize() {
  tokens[0] = &buffer[1];
  char *s = &buffer[2];
  tokenCnt = 1;
  uint8_t achkSum = ESRAPI_SOC + buffer[1];
  uint8_t xchkSum = ESRAPI_SOC ^ buffer[1];
  uint8_t hchkSum;
  uint8_t chktype=0; // 0=none,1=additive,2=xor
  while (*s) {
    if (*s == ' ') {
      achkSum += *s;
      xchkSum ^= *s;
      *s = '\0';
      tokens[tokenCnt++] = ++s;}
    else if ((*s == '*') ||// additive checksum
	     (*s == '^')) { // XOR checksum
      if (*s == '*') chktype = 1;
      else if (*s == '^') chktype = 2;
      *(s++) = '\0';
      hchkSum = htou8(s);
      break;}
    else {
      achkSum += *s;
      xchkSum ^= *(s++);}}
  return ((chktype == 0) ||
	  ((chktype == 1) && (hchkSum == achkSum)) ||
	  ((chktype == 2) && (hchkSum == xchkSum))) ? 0 : 1;}

int EvseRapiProcessor::processCmd() {
  Serial.println(buffer);
  UNION4B u1,u2,u3;
  int rc = 0;
  // we use bufCnt as a flag in response() to signify data to write
  bufCnt = 0;
  cmdIdCnt = 0;
  char *s = tokens[0];
  switch(*(s++)) {
  //+++ F +++ FUNCTION ++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  case 'F': // function
    switch(*s) {
    case 'A': // Allow charging
      g_flagAllowCharging = atoi(EVCS_BUS_STR)*10+atoi(EVCS_ID_STR); //AC_Flag = BusID
      sprintf(g_sTmp,"%5d",g_flagAllowCharging);
      Serial.println("Charging allowed"); 
      break;
    case 'C': // Cancel Charging
      g_flagAllowCharging = 0;
      sprintf(g_sTmp,"%5d",g_flagAllowCharging);
      Serial.println("Charging canceled"); 
      break;
    case 'D': // disable EVSE
      Serial.println("EVCS DISABLED");
      digitalWrite(LED_BUILTIN, LOW);      
      sleep_flag = 0;
      break;
    case 'E': // enable EVSE
      Serial.println("EVCS ENABLED");
      digitalWrite(LED_BUILTIN, HIGH);
      sleep_flag = 0;
      break;
    case 'P':  // print to LCD
      u1.u = dtou32(tokens[1]); // x
      u2.u = dtou32(tokens[2]); // y
      // now restore the spaces that were replaced w/ nulls by tokenizing
      for (u3.i=4;u3.i < tokenCnt;u3.i++) {
        *(tokens[u3.i]-1) = ' ';
      }
      Serial.print("'");
      Serial.print(tokens[3]);
      Serial.print("' on display x = ");
      Serial.print(u1.u);
      Serial.print(" y = ");
      Serial.println(u2.u);      
      break;
    
    case 'R': // reset EVSE
      Serial.println("EVCS RESET");
      digitalWrite(LED_BUILTIN, LOW);
      delay(100);
      digitalWrite(LED_BUILTIN, HIGH);
      delay(100);
      digitalWrite(LED_BUILTIN, LOW);
      delay(100);
      digitalWrite(LED_BUILTIN, HIGH);      
      sleep_flag = 0;
      break;
    case 'S': // sleep
      Serial.println("EVCS SLEEPING");
      sleep_flag = 1;
      break;
    default:
      rc = -1; // unknown
      break;
    }
    break;
  //+++ S +++ SET PARAMETER++++++++++++++++++++++++++++++++++++++++++++++++++++
  case 'S': // set parameter
    switch(*s) {
    case '1': // set RTC
      if (tokenCnt == 7) {        
        setTime(dtou32(tokens[4]),dtou32(tokens[5]),dtou32(tokens[6]),dtou32(tokens[3]),dtou32(tokens[2]),dtou32(tokens[1]));
        sprintf(out_buffer,"%d:%d:%d %d/%d/%d", hour(), minute(), second(), day(), month(), year());      
        Serial.print("Clock set: ");
        Serial.println(out_buffer);
      }
      break;
    case '3': // set time limit
      if (tokenCnt == 2) {
        time_limit = dtou32(tokens[1]);  
        Serial.print("Setting charge time limit to: ");
        Serial.print(time_limit*15);
        Serial.println(" minutes");
      }
      break;
    case 'C': // current capacity
      if (tokenCnt == 2) {
        current_capacity = dtou32(tokens[1]);  
        if (current_capacity > MAX_CURRENT_CAPACITY) {
          current_capacity = MAX_CURRENT_CAPACITY;
          Serial.print("Error, Value too high! Current capacity set to Max = ");
          Serial.print(current_capacity);
          Serial.println(" A");
          rc = 1;
        }
        else if (current_capacity < MIN_CURRENT_CAPACITY) {
          current_capacity = MIN_CURRENT_CAPACITY;
          Serial.print("Error, Value too low! Current capacity set to Min = ");
          Serial.print(current_capacity);
          Serial.println(" A");
          rc = 1;
        }
        else {
          Serial.print("Setting current capcity to: ");
          Serial.print(current_capacity);
          Serial.println(" A");
        rc = 0;
        }        
      }
      break;
    case 'D': // diode check
      if (tokenCnt == 2) {
        diode_check = dtou32(tokens[1]);          
        if (diode_check) {
          Serial.println("Diode check is on");
        }
        else {
          Serial.println("Diode check is off");
        }
      }  
      break;
/*    case 'E': // echo  //blocking the whole program??
      if (tokenCnt == 2) {
        echo = (byte)dtou32(tokens[1]);          
        Serial.print("Echo: ");
        Serial.println(echo,HEX);
        if (echo) {
          Serial.println("RAPI echo is on");
        }
        else {
          Serial.println("RAPI echo is off");
        }
      }  
      break;*/
    case 'G': // ground check
      if (tokenCnt == 2) {
        ground_check = dtou32(tokens[1]);
        if (ground_check) {
          Serial.println("Ground check is on");
        }
        else {
          Serial.println("Ground check is off");
        }
      }  
      break;
    case 'H': // cHarge limit      
      if (tokenCnt == 2) {
        charge_limit = dtou32(tokens[1]);
        Serial.print("Charge limit set to ");
        Serial.print(charge_limit);
        Serial.println(" kWh");
      }
      break;
    case 'O':      
      if (tokenCnt == 3) {
        temp_ambient_thresh = dtou32(tokens[1]);
        //temp_ir_thresh = dtou32(tokens[2]);        
        Serial.print("Setting Overtemp threshold to ");        
        sprintf(out_buffer,"%d/10°C",int(temp_ambient_thresh));//,int(temp_ir_thresh));        
        Serial.println(out_buffer);      
      }
      break;
    case 'R': // stuck relay check      
      if (tokenCnt == 2) {
        stuck_relay_check = dtou32(tokens[1]);
        if (stuck_relay_check) {
          Serial.println("Stuck relay check is on");
        }
        else {
          Serial.println("Stuck relay check is off");
        }        
      }
      break;
    case 'T': // timer      
      if (tokenCnt == 5) {
        //extern DelayTimer g_DelayTimer;
        if ((*tokens[1] == '0') && (*tokens[2] == '0') && (*tokens[3] == '0') && (*tokens[4] == '0')) {
          Serial.print("Delay timer is disabled");        
        }
        else {
          Serial.print("Start timer set to hour: ");        
          sprintf(out_buffer,"%d:%d",(int)dtou32(tokens[1]), (int)dtou32(tokens[2]));        
          Serial.println(out_buffer);
          Serial.print("Stop timer set to hour: ");        
          sprintf(out_buffer,"%d:%d",(int)dtou32(tokens[3]), (int)dtou32(tokens[4]));             
          Serial.println(out_buffer);          
        }
      }
      break;
    default:
      rc = -1; // unknown
      break;
    }
    break;
  //+++ G +++ GET PARAMETER +++++++++++++++++++++++++++++++++++++++++++++++++++
  case 'G': // get parameter
    switch(*s) {
    case '3': // get time limit
      sprintf(buffer,"%d",(byte)time_limit);
      Serial.print("Charge time limit is set to:");
      Serial.print(time_limit*15);
      Serial.println(" minutes");
      bufCnt = 1; // flag response text output
      break;
    case 'C': // get current capacity range
      Serial.print("Current capacity range goes from ");
      Serial.print(MIN_CURRENT_CAPACITY);
      Serial.print("A to ");
      Serial.print(MAX_CURRENT_CAPACITY);
      Serial.println("A");
      u1.i = MIN_CURRENT_CAPACITY;
      u2.i = MAX_CURRENT_CAPACITY;
      sprintf(buffer,"%d %d",u1.i,u2.i);
      bufCnt = 1; // flag response text output
      break;
    case 'E': // get settings
      u1.u = DEFAULT_CURRENT_CAPACITY;
      u2.u = m_wFlags;
      Serial.print("Default current capacity is ");
      Serial.print(u1.u);
      Serial.print(" A\nFlags in HEX: 0x");
      Serial.println(u2.u,HEX);
      sprintf(buffer,"%d %04x",u1.u,u2.u);
      bufCnt = 1; // flag response text output
      break;
    case 'F': // get fault counters
      u1.u = 0;
      u2.u = NoGndTripCnt;
      u3.u = StuckRelayTripCnt;
      Serial.print("Trip counters: RCD, noGND, stuck relay in HEX: 0x");      
      Serial.print(u1.u,HEX);
      Serial.print(" 0x");
      Serial.print(u2.u,HEX);
      Serial.print(" 0x");
      Serial.println(u3.u,HEX);      
      sprintf(buffer,"%x %x %x",u1.u,u2.u,u3.u);
      bufCnt = 1; // flag response text output
      break;
    case 'G':
      u1.i32 = 12458;
      u2.i32 = 234125;
      Serial.print("Charging current is: ");      
      Serial.print(u1.i32);
      Serial.print(" mA\nCharging voltage is: ");
      Serial.print(u2.i32);
      Serial.println(" mV");
      sprintf(buffer,"%ld %ld",u1.i32,u2.i32);
      bufCnt = 1; // flag response text output
      break;
    case 'H': // get cHarge limit    
      Serial.print("Charge limit in kWh: ");      
      Serial.println(charge_limit);
      sprintf(buffer,"%d",(int)charge_limit);
      bufCnt = 1; // flag response text output
      break;
    case 'O':
      Serial.print("Overtemperature threshhold in 10th of °C: ");      
      Serial.println(temp_ambient_thresh);
      sprintf(buffer,"%d",(int)temp_ambient_thresh);
      bufCnt = 1; // flag response text output
      break;
    case 'P':
      Serial.print("Temperature 10th of °C: ");      
      Serial.println(temp);
      sprintf(buffer,"%d",(int)temp);
      bufCnt = 1; // flag response text output
      break;
    case 'R': // get RFID NUID
      Serial.println("RFID NUID is 11324877");      
      sprintf(buffer,"%02x%02x%02x%02x",11,32,48,77);
        bufCnt = 1; // flag response text output
      break;
    case 'S': // get state
      ElapsedChargeTime= millis()/1000;
      Serial.print("State is ");
      switch (evcs_state) {
       case 0:
         Serial.print("unknown");
         break;
       case 1:
         Serial.print("A (12V) - EV not connected");
         break;
       case 2:
         Serial.print("B (9V) - EV connected and ready");
         break;
       case 3:
         Serial.print("C (6V) - EV charging");
         break;
      }
      Serial.print("\n Elapsed charge time in seconds: ");           
      Serial.println(ElapsedChargeTime);           
      sprintf(buffer,"%d %ld",evcs_state,ElapsedChargeTime);
      bufCnt = 1; // flag response text output
      break;
    case 'T': // get time
      sprintf(buffer,"%d %d %d %d %d %d",(year()-2000), month(), day(), hour(), minute(), second());      
      Serial.print("Date and time (year, month, day, hour, minutes, seconds): ");      
      Serial.println(buffer);      
      bufCnt = 1; // flag response text output
      break;
    case 'U':
      g_WattSeconds = 2916*millis()/1000;
      g_WattHours_accumulated = 235+(g_WattSeconds/3600);
      Serial.print("Energy: ");      
      Serial.print(g_WattSeconds);      
      Serial.print(" Ws\nAccumulated Energy: ");      
      Serial.print(g_WattHours_accumulated);
      Serial.println(" kWh");      
      sprintf(buffer,"%lu %lu",g_WattSeconds,g_WattHours_accumulated);
      bufCnt = 1;
      break;
    case 'V': // get version
      GetVerStr(buffer);      
      strcat(buffer," ");
      strcat_P(buffer,RAPI_VER);
      Serial.print("Version: ");      
      Serial.println(buffer);      
      bufCnt = 1; // flag response text output
      break;
    default:
      rc = -1; // unknown
      break;
    }
    break;
  //+++ DEFAULT +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  default:
    rc = -1; // unknown
    break;
  }

  response((rc == 0) ? 1 : 0);

  reset();

  return rc;
}

void EvseRapiProcessor::response(uint8_t ok) {  
  write(ESRAPI_SOC);
  write(EVCS_BUS);
  write(EVCS_ID);
  write(ok ? "OK " : "NK ");
  if (bufCnt) {
    write(buffer);
  }
  write(ESRAPI_EOC);
  //if (echo) write('\n'); 
}

EvseSerialRapiProcessor::EvseSerialRapiProcessor() {}

void EvseSerialRapiProcessor::init() {EvseRapiProcessor::init();}

EvseSerialRapiProcessor g_ESRP;

void RapiInit() {g_ESRP.init();}

void RapiDoCmd() {  g_ESRP.doCmd();}

void RapiSendEvseState(uint8_t nodupe){
  static uint8_t evseStateSent = EVCS_STATE_UNKNOWN;
  uint8_t es = EVCS_STATE_B;
  if (!nodupe || (evseStateSent != es)) {
    g_ESRP.sendEvseState();
    evseStateSent = es;
  }
}

void RapiSendRFID_NUID(uint8_t nodupe){
//  static uint8_t evseStateSent = EVCS_STATE_UNKNOWN;
//  uint8_t es = g_EvseController.GetState();
//  if (!nodupe || (evseStateSent != es)) {
    g_ESRP.sendRFID_NUID();
//    evseStateSent = es;
//  }
}
